// FtController.java

/*
This software is part of the Aplu FtController package.
It is Open Source Free Software, so you may
- run the code for any purpose
- study how the code works and adapt it to your needs
- integrate all or parts of the code in your own programs
- redistribute copies of the code
- improve the code and release your improvements to the public
However the use of the code is entirely your responsibility.
 */

package ch.aplu.ftcontroller;

import ch.aplu.jaw.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.*;

/**
 * Class to handle a fischertechnik ROBO interface connected directly to
 * the USB port or using the wireless RF Data Link.
 * A FtController instance communicates with the umFish40.dll using
 * the Java API Wrapper (JAW) framework based on the Java Native Interface (JNI).
 * Any changes on input ports are considered as events. A high performance
 * native thread polls the interface and reports changes to a Java thread that
 * fires listener callback methods that conform to the Java event model.
 * All listener callbacks run in the Event Dispatch Thread EDT,
 * so it is safe to use Swing methods directly.
 */
public class FtController
{
  private class MyActionListener implements ActionListener
  {
    public void actionPerformed(ActionEvent e)
    {
      ControllerState cs = readData();

      if (cs != null && ((cs.source & EVT_CONNECT) != 0))
      {
        isConnected = false;
        wasDisconnected = true;
        if (connectionListener != null)
          connectionListener.connectionLost();
      }

      for (int i = 0; i < 4; i++)
      {
        if (endSwitchListeners[i] != null)
        {
          if (cs != null && ((cs.source & EVT_ESWITCH) != 0))
          {
            int value = cs.endSwitch;
            int mask = 1 << i;
            if ((value & mask) != 0)
              endSwitchListeners[i].isOn();
          }
        }
      }

      if (digitalListener != null)
      {
        if (cs != null && ((cs.source & EVT_DIN) != 0))
        {
          int value = cs.digitalIn & 0x0000FFFF;
          int mask = cs.digitalIn >> 16;
          DigitalEvent evt = new DigitalEvent();
          evt.value = value;
          evt.mask = mask;
          digitalListener.digitalChanged(evt);
        }
      }

      for (int i = 0; i < 2; i++)
      {
        if (resistorListeners[i] != null)
        {
          if (cs != null && ((cs.source & EVT_RIN[i]) != 0))
          {
            ResistorEvent evt = new ResistorEvent();
            evt.port = i + 1;
            evt.value = cs.resistorIn[i];
            resistorListeners[i].resistorChanged(evt);
          }
        }
      }

      for (int i = 0; i < 2; i++)
      {
        if (analogListeners[i] != null)
        {
          if (cs != null && ((cs.source & EVT_AIN[i]) != 0))
          {
            AnalogEvent evt = new AnalogEvent();
            evt.port = i + 1;
            evt.value = cs.analogIn[i];
            analogListeners[i].analogChanged(evt);
          }
        }
      }

      for (int i = 0; i < 2; i++)
      {
        if (distanceListeners[i] != null)
        {
          if (cs != null && ((cs.source & EVT_UIN[i]) != 0))
          {
            DistanceEvent evt = new DistanceEvent();
            evt.port = i + 1;
            evt.value = cs.distanceIn[i];
            distanceListeners[i].distanceChanged(evt);
          }
        }
      }

      for (int i = 0; i < 8; i++)
      {
        if (counterListeners[i] != null)
        {
          if (cs != null && ((cs.source & EVT_COUNTER[i]) != 0))
          {
            CounterEvent evt = new CounterEvent();
            evt.counterNb = i + 1;
            evt.count = cs.counter[i];
            counterListeners[i].countChanged(evt);
          }
        }
      }

      if (infraredListener != null)
      {
        if (cs != null && ((cs.source & EVT_IRIN) != 0))
        {
          InfraredEvent evt = new InfraredEvent();
          evt.value = cs.irIn;
          infraredListener.infraredChanged(evt);
        }
      }
    }
  }
  // ----------------- End of inner class --------------

  private static final String version = "1.03";
  private static final int defaultControllerPollPeriod = 20;  // ms
  private static final int defaultQueuePollPeriod = 20;  // ms
  private String firmwareInfo;
  private String deviceName;
  protected final int[] EVT_COUNTER =
  {
    0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080
  };
  protected final int EVT_DIN = 0x0100;
  protected final int[] EVT_RIN =
  {
    0x0200, 0x0400
  };
  protected final int[] EVT_AIN =
  {
    0x0800, 0x1000
  };
  protected final int[] EVT_UIN =
  {
    0x2000, 0x4000
  };

  protected static final int EVT_IRIN = 0x8000;
  protected static final int EVT_ESWITCH = 0x10000;
  protected static final int EVT_CONNECT = 0x20000;
  private int error = (int)0xE0000001L;
  /**
   * A reference to the motor 1 instance.
   */
  public Motor motor1 = new Motor(this, 1);
  /**
   * A reference to the motor 2 instance.
   */
  public Motor motor2 = new Motor(this, 2);
  /**
   * A reference to the motor 3 instance.
   */
  public Motor motor3 = new Motor(this, 3);
  /**
   * A reference to the motor 4 instance.
   */
  public Motor motor4 = new Motor(this, 4);
  /**
   * Array of all motors instances.
   * Index 0 corresponds to motor 1.
   */
  public Motor[] motorAry =
  {
    motor1, motor2, motor3, motor4
  };
  /**
   * A reference to the counter 1 instance.
   */
  public Counter counter1 = new Counter(this, 1);
  /**
   * A reference to the counter 2 instance.
   */
  public Counter counter2 = new Counter(this, 2);
  /**
   * A reference to the counter 3 instance.
   */
  public Counter counter3 = new Counter(this, 3);
  /**
   * A reference to the counter 4 instance.
   */
  public Counter counter4 = new Counter(this, 4);
  /**
   * A reference to the counter 5 instance.
   */
  public Counter counter5 = new Counter(this, 5);
  /**
   * A reference to the counter 6 instance.
   */
  public Counter counter6 = new Counter(this, 6);
  /**
   * A reference to the counter 7 instance.
   */
  public Counter counter7 = new Counter(this, 7);
  /**
   * A reference to the counter 8 instance.
   */
  public Counter counter8 = new Counter(this, 8);
  /**
   * Array of all counter instances.
   * Index 0 corresponds to counter 1.
   */
  public Counter counterAry[] =
  {
    counter1, counter2, counter3, counter4,
    counter5, counter6, counter7, counter8
  };
  /**
   * A reference to the digitalIn instance.
   */
  public DigitalIn digitalIn = new DigitalIn(this);
  /**
   * A reference to the analogIn 1 instance.
   */
  public AnalogIn analogIn1 = new AnalogIn(this, 1);
  /**
   * A reference to the analogIn 2 instance.
   */
  public AnalogIn analogIn2 = new AnalogIn(this, 2);
  /**
   * Array of all analogIn instances.
   * Index 0 corresponds to analogIn 1.
   */
  public AnalogIn[] analogInAry =
  {
    analogIn1, analogIn2
  };
  /**
   * A reference to the resistorIn 1 instance.
   */
  public ResistorIn resistorIn1 = new ResistorIn(this, 1);
  /**
   * A reference to the resistorIn 2 instance.
   */
  public ResistorIn resistorIn2 = new ResistorIn(this, 2);
  /**
   * Array of all resistorIn instances.
   * Index 0 corresponds to resistorIn 1.
   */
  public ResistorIn[] resistorInAry =
  {
    resistorIn1, resistorIn2
  };
  /**
   * A reference to the distanceIn 1 instance.
   */
  public DistanceIn distanceIn1 = new DistanceIn(this, 1);
  /**
   * A reference to the distanceIn 2 instance.
   */
  public DistanceIn distanceIn2 = new DistanceIn(this, 2);
  /**
   * Array of all distanceIn instances.
   * Index 0 corresponds to distanceIn 1.
   */
  public DistanceIn[] distanceInAry =
  {
    distanceIn1, distanceIn2
  };
  /**
   * A reference to the infraredIn instance.
   */
  public InfraredIn infraredIn = new InfraredIn(this);
  /**
   * A reference to the lamp 1 instance.
   */
  public Lamp lamp1 = new Lamp(this, 1);
  /**
   * A reference to the lamp 2 instance.
   */
  public Lamp lamp2 = new Lamp(this, 2);
  /**
   * A reference to the lamp 3 instance.
   */
  public Lamp lamp3 = new Lamp(this, 3);
  /**
   * A reference to the lamp 4 instance.
   */
  public Lamp lamp4 = new Lamp(this, 4);
  /**
   * A reference to the lamp 5 instance.
   */
  public Lamp lamp5 = new Lamp(this, 5);
  /**
   * A reference to the lamp 6 instance.
   */
  public Lamp lamp6 = new Lamp(this, 6);
  /**
   * A reference to the lamp 7 instance.
   */
  public Lamp lamp7 = new Lamp(this, 7);
  /**
   * A reference to the lamp 8 instance.
   */
  public Lamp lamp8 = new Lamp(this, 8);
  /**
   * Array of all lamp instances.
   * Index 0 corresponds to lamp 1.
   */
  public Lamp lampAry[] =
  {
    lamp1, lamp2, lamp3, lamp4,
    lamp5, lamp6, lamp7, lamp8
  };
  private int[] values = new int[19];
  private int queuePollPeriod; // Polling period of the Java thread
  private int controllerPollPeriod;  // Polling period of native thread in ms
  protected NativeHandler nh;
  protected int motorNb;
  protected int lampNb;
  protected int counterNb;
  protected int count;
  protected int[] speeds =
  {
    8, 8, 8, 8
  };
  protected int[] intensities =
  {
    8, 8, 8, 8, 8, 8, 8, 8
  };
  protected boolean isLampOn;
  protected int[] steps = new int[4];
  protected boolean isConnected = false;
  private boolean wasDisconnected = false;
  private int eventMask;
  private DigitalListener digitalListener = null;
  private AnalogListener[] analogListeners;
  private ResistorListener[] resistorListeners;
  private DistanceListener[] distanceListeners;
  private CounterListener[] counterListeners;
  private CounterListener[] saveCounterListeners;
  private EndSwitchListener[] endSwitchListeners;
  protected MotorListener[] motorListeners;
  protected MotorListener[] saveMotorListeners;
  private ConnectionListener connectionListener;
  protected boolean[] enableEndSwitchReport;
  protected int a1Jitter = 0;
  protected int a2Jitter = 0;
  protected int axJitter = 0;
  protected int ayJitter = 0;
  private InfraredListener infraredListener = null;
  private Timer pollTimer = null;
  private Object monitor = new Object();

  /**
   * Same as FtController("ftcontroller", true, 20, 20)
   * @see #FtController(String dll, boolean doConnect, int controllerPollPeriod, int queuePollPeriod)
   */
  public FtController()
  {
    this("ftcontroller", true, defaultControllerPollPeriod, defaultQueuePollPeriod);
  }

  /**
   * Same as FtController("ftcontroller", doConnect, 20, 20)
   * @see #FtController(String dll, boolean doConnect, int controllerPollPeriod, int queuePollPeriod)
   */
  public FtController(boolean doConnect)
  {
    this("ftcontroller", doConnect, defaultControllerPollPeriod, defaultQueuePollPeriod);
  }

  /**
   * Same as FtController(dll, true, 20, 20)
   * @see #FtController(String dll, boolean doConnect, int controllerPollPeriod, int queuePollPeriod)
   */
  public FtController(String dll)
  {
    this(dll, true, defaultControllerPollPeriod, defaultQueuePollPeriod);
  }

  /**
   * Same as FtController(dll, doConnect, 20, 20)
   * @see #FtController(String dll, boolean doConnect, int controllerPollPeriod, int queuePollPeriod)
   */
  public FtController(String dll, boolean doConnect)
  {
    this(dll, doConnect, defaultControllerPollPeriod, defaultQueuePollPeriod);
  }

  /**
   * Creates a FtController instance using the given Windows dll.
   * If doConnect = true, tries to establish a connection to the controller using connect(). If it
   * fails, an error message is displayed and when closed the application terminates.
   * If you want to handle the connection error yourself, set doConnect = false and check
   * the return value when calling connect().
   * @param dll either the fully qualified filename to the Windows DLL or simply
   * the name of the DLL without extension. In this case the DLL must be found
   * in the operating system's path.
   * @param doConnect if true, tries to establish a connection
   * @param controllerPollPeriod the period in ms for the native polling thread
   * @param queuePollPeriod the period in ms for the Java polling thread
   * @see #connect()
   */
  public FtController(String dll, boolean doConnect, int controllerPollPeriod, int queuePollPeriod)
  {
    this.controllerPollPeriod = controllerPollPeriod;
    this.queuePollPeriod = queuePollPeriod;
    analogListeners = new AnalogListener[2];
    for (int i = 0; i < 2; i++)
      analogListeners[i] = null;
    resistorListeners = new ResistorListener[2];
    for (int i = 0; i < 2; i++)
      resistorListeners[i] = null;
    distanceListeners = new DistanceListener[2];
    for (int i = 0; i < 2; i++)
      distanceListeners[i] = null;
    counterListeners = new CounterListener[8];
    for (int i = 0; i < 8; i++)
      counterListeners[i] = null;
    saveCounterListeners = new CounterListener[8];
    for (int i = 0; i < 8; i++)
      saveCounterListeners[i] = null;
    motorListeners = new MotorListener[4];
    for (int i = 0; i < 4; i++)
      motorListeners[i] = null;
    saveMotorListeners = new MotorListener[4];
    for (int i = 0; i < 4; i++)
      saveMotorListeners[i] = null;
    endSwitchListeners = new EndSwitchListener[4];
    for (int i = 0; i < 4; i++)
      endSwitchListeners[i] = null;
    enableEndSwitchReport = new boolean[4];
    for (int i = 0; i < 4; i++)
      enableEndSwitchReport[i] = false;
    nh = new NativeHandler(dll, this);
    nh.createBuf(values);  // JNIBuffer with default bufSize

    if (doConnect)
    {
      boolean rc = connect();
      if (!rc)
      {
        JOptionPane.showMessageDialog(null, "Connection to the controller failed",
          "Fatal error",
          JOptionPane.ERROR_MESSAGE);
        nh.destroy();
        System.exit(1);
      }
    }
  }

  /**
   * Establishes a connection to the controller with USB or RF Data Link
   * by calling the native function rbOpenInterfaceUSBdis(0, 0, ftiDisUltra)
   * (ports D1/D2 used for ultrasonic distance sensors).
   * After the connection is lost, you cannot reestablish it by
   * calling connect, you must restart the program.
   * @return true if successful or already connected; false, if connection fails
   */
  public boolean connect()
  {
    if (isConnected)
      return true;
    if (wasDisconnected)
      return false;
    int rc = nh.invoke(0);
    if (rc == error)
      return false;

    nh.startThread();      // Start high priority thread

    while (nh.invoke(100) == 0)  // Wait until thread is ready
      delay(50);

    isConnected = true;
    return true;
  }

  /**
   * Releases all resources.
   * You should call dispose in any case before quitting the program.
   */
  public void dispose()
  {
    if (!isConnected)
      return;
    if (pollTimer != null)
      pollTimer.stop();
    digitalIn.release();
    delay(1000);  // Wait a moment gives better stop behaviour, not clear why
    nh.destroy();
  }

  /**
   * Releases all resources and call System.exit(0).
   */
  public void exit()
  {
    dispose();
    System.exit(0);
  }

  /**
   * Delay the execution for the given time.
   * @param ms the wait time in milliseconds
   */
  public static void delay(int ms)
  {
    try
    {
      Thread.currentThread().sleep(ms);
    }
    catch (InterruptedException ex)
    {
    }
  }

  /**
   * Set all counters to zero.
   */
  public synchronized void clearAllCounters()
  {
    if (!isConnected)
      return;
    nh.invoke(9);
  }

  /**
   * Returns the device type.
   * @return the active device type, -1 if not connected
   */
  public synchronized int getDeviceType()
  {
    if (!isConnected)
      return -1;
    return nh.invoke(17);
  }

  /**
   * Returns the device serial number.
   * @return the active device serial number, -1 if not connected
   */
  public synchronized int getSerialNumber()
  {
    if (!isConnected)
      return -1;
    return nh.invoke(18);
  }

  /**
   * Returns the firmware number.
   * @return the active device firmware number , -1 if not connected
   */
  public synchronized int getFirmwareNumber()
  {
    if (!isConnected)
      return -1;
    return nh.invoke(19);
  }

  /**
   * Returns the firmware information.
   * @return the active device firmware information, empty if not connected
   */
  public synchronized String getFirmwareInfo()
  {
    if (!isConnected)
      return "";
    nh.invoke(20);
    return firmwareInfo;
  }

  /**
   * Returns the device name.
   * @return the active device name, empty if not connected
   */
  public synchronized String getDeviceName()
  {
    if (!isConnected)
      return "";
    nh.invoke(21);
    return deviceName;
  }

  /**
   * Returns status of all motors as a bit mask. Request is made to
   * the hardware controller.
   * Motor 1 left: b0<br>
   * Motor 1 right: b1<br>
   * Motor 2 left: b2<br>
   * Motor 2 right: b3<br>
   * Motor 3 left: b4<br>
   * Motor 3 right: b5<br>
   * Motor 4 left: b6<br>
   * Motor 4 right: b7<br>
   */
  public synchronized int getMotorsStatus()
  {
    return nh.invoke(24);
  }

  /**
   * Registers the given DigitalListener.
   * Starts polling the event queue if not yet done for another listener.
   * For the isPortChanged() callback, the initializing port state
   * corresponds to the state when this or the connect method is called.
   * @param digitalListener the DigitalListener to register
   */
  public void addDigitalListener(DigitalListener digitalListener)
  {
    if (digitalListener == null)
      return;

    if (isConnected)
      synchronized (this)
      {
        nh.invoke(102); // Set default digitalIn
      }
    eventMask =
      eventMask | EVT_DIN;
    synchronized (this)
    {
      nh.invoke(103);
    }
    this.digitalListener = digitalListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Registers the given AnalogListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param analogListener the AnalogListener to register
   * @param portNb the port number to monitor (1, 2)
   * @param jitter the dead zone around the current value that must be exceeded to report an event
   */
  public void addAnalogListener(AnalogListener analogListener,
    int portNb, int jitter)
  {
    if (analogListener == null || portNb < 1 || portNb > 2)
      return;

    if (portNb == 1)
      a1Jitter = jitter;
    if (portNb == 2)
      a2Jitter = jitter;
    addAnalogListener(analogListener, portNb);
  }

  /**
   * Same as addAnalogListener(analogListener, portNb, 0).
   */
  public void addAnalogListener(AnalogListener analogListener, int portNb)
  {
    if (analogListener == null || portNb < 1 || portNb > 2)
      return;

    eventMask = eventMask | EVT_AIN[portNb - 1];
    synchronized (this)
    {
      nh.invoke(103);
    }
    analogListeners[portNb - 1] = analogListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Registers the given ResitorListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param resistorListener the ResistorListener to register
   * @param portNb the port number to monitor (1, 2)
   * @param jitter the dead zone around the current value that must be exceeded to report an event
   */
  public void addResistorListener(ResistorListener resistorListener,
    int portNb, int jitter)
  {
    if (resistorListener == null || portNb < 1 || portNb > 2)
      return;

    if (portNb == 1)
      axJitter = jitter;
    if (portNb == 2)
      ayJitter = jitter;
    addResistorListener(resistorListener, portNb);
  }

  /**
   * Same as addResistorListener(resistorListener, portNb, 0).
   */
  public void addResistorListener(ResistorListener resistorListener,
    int portNb)
  {
    if (resistorListener == null || portNb < 1 || portNb > 2)
      return;

    eventMask = eventMask | EVT_RIN[portNb - 1];
    synchronized (this)
    {
      nh.invoke(103);
    }
    resistorListeners[portNb - 1] = resistorListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Registers the given DistanceListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param distanceListener the DistanceListener to register
   * @param portNb the port number to monitor (1, 2)
   */
  public void addDistanceListener(DistanceListener distanceListener,
    int portNb)
  {
    if (distanceListener == null || portNb < 1 || portNb > 2)
      return;

    eventMask = eventMask | EVT_UIN[portNb - 1];
    synchronized (this)
    {
      nh.invoke(103);
    }
    distanceListeners[portNb - 1] = distanceListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Registers the given InfraredListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param infraredListener the InfraredListener to register
   */
  public void addInfraredListener(InfraredListener infraredListener)
  {
    if (infraredListener == null)
      return;

    eventMask = eventMask | EVT_IRIN;
    synchronized (this)
    {
      nh.invoke(103);
    }
    this.infraredListener = infraredListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  private void addEndSwitchListener(EndSwitchListener endSwitchListener,
    int switchNb)
  {
    if (endSwitchListener == null || switchNb < 1 || switchNb > 4)
      return;

    eventMask = eventMask | EVT_ESWITCH;
    synchronized (this)
    {
      nh.invoke(103);
    }
    this.endSwitchListeners[switchNb - 1] = endSwitchListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Registers the given CounterListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param counterListener the CounterListener to register
   * @param counterNb the counter number to monitor (1..8)
   */
  public void addCounterListener(CounterListener counterListener, int counterNb)
  {
    if (counterListener == null || counterNb < 1 || counterNb > 8)
      return;

    eventMask = eventMask | EVT_COUNTER[counterNb - 1];
    synchronized (this)
    {
      nh.invoke(103);
    }
    counterListeners[counterNb - 1] = counterListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Same as addMotorListener(MotorListener motorListener, int motorNb) for
   * given motor.
   */
  public void addMotorListener(MotorListener motorListener, final Motor motor)
  {
    addMotorListener(motorListener, motor.getNumber());
  }

  /**
   * Registers the given MotorListener for the given motor.
   * Starts polling the event queue if not yet done for another listener.
   * The listener assumes that the motor will run in RobMode, where some
   * of the counters are used to for this purpose (more information 
   * in class Motor).
   * @param motorListener the MotorListener to register
   * @param motorNb the number of the motor
   * @see ch.aplu.ftcontroller.Motor
   */
  public void addMotorListener(MotorListener motorListener, final int motorNb)
  {
    addMotorListener(motorListener, motorNb, true);
  }

  protected void addMotorListener(MotorListener motorListener, final int motorNb,
    final boolean saveCounter)
  {
    if (motorListener == null || motorNb < 1 || motorNb > 4)
      return;
    final int motorIndex = motorNb - 1;
    motorListeners[motorIndex] = motorListener;
    final int counterNb = 2 * motorNb;
    if (saveCounter)
      saveCounterListeners[counterNb - 1] = counterListeners[counterNb - 1];

    addEndSwitchListener(new EndSwitchListener()
    {
      public void isOn()
      {
        if (motorListeners[motorIndex] != null &&
          enableEndSwitchReport[motorIndex])
        {
          motorListeners[motorIndex].motorStopped(motorNb, true);
          if (saveCounter && saveCounterListeners[counterNb - 1] != null)
            saveCounterListeners[counterNb - 1].countChanged(new CounterEvent(counterNb, 0));
          motorAry[motorIndex].setRotating(false);
        }
      }
    }, motorNb);

    addCounterListener(new CounterListener()
    {
      public void countChanged(CounterEvent evt)
      {
        if (saveCounter && saveCounterListeners[counterNb - 1] != null)
          saveCounterListeners[counterNb - 1].countChanged(evt);

        if (!motorAry[motorIndex].isRotating())
          return;
        int value = evt.getCount();  // Counts backward

        int curSteps = steps[motorIndex] - value;
        if (motorListeners[motorIndex] != null)
        {
          if (curSteps != 0)
            motorListeners[motorIndex].stepChanged(motorNb, curSteps);

          int motorStatus = getMotorsStatus();
          int leftBitMask = (1 << 2 * motorIndex);
          int rightBitMask = (1 << 2 * motorIndex + 1);

          if ((motorStatus & leftBitMask) == 0 && (motorStatus & rightBitMask) == 0)
          {
            motorListeners[motorIndex].motorStopped(motorNb, false);
            motorAry[motorIndex].setRotating(false);
          }
        }
      }
    }, counterNb);
  }

  protected void saveMotorListener(Motor motor)
  {
    int motorNb = motor.getNumber();
    saveMotorListeners[motorNb - 1] = motorListeners[motorNb - 1];
  }

  protected void restoreMotorListener(Motor motor)
  {
    int motorNb = motor.getNumber();
    motorListeners[motorNb - 1] = saveMotorListeners[motorNb - 1];
    saveMotorListeners[motorNb - 1] = null;
  }

  /**
   * Registers the given ConnectionListener.
   * Starts polling the event queue if not yet done for another listener.
   * @param connectionListener the ConnectionListener to register
   */
  public void addConnectionListener(ConnectionListener connectionListener)
  {
    if (connectionListener == null)
      return;

    this.connectionListener = connectionListener;
    if (pollTimer == null)
    {
      pollTimer = new Timer(queuePollPeriod, new MyActionListener());
      pollTimer.start();
    }
  }

  /**
   * Returns true, if connected; otherwise false.
   * @return Connection state
   */
  public boolean isConnected()
  {
    return isConnected;
  }

  private synchronized ControllerState readData()
  {
    if (nh.countBuf() <= 0)
      return null;
    nh.readBuf(19);

    ControllerState cs = new ControllerState();
    for (int i = 0; i < 8; i++)
      cs.counter[i] = values[i];
    cs.digitalIn = values[8];
    cs.resistorIn[0] = values[9];
    cs.resistorIn[1] = values[10];
    cs.analogIn[0] = values[11];
    cs.analogIn[1] = values[12];
    cs.distanceIn[0] = values[13];
    cs.distanceIn[1] = values[14];
    cs.irIn = values[15];
    cs.endSwitch = values[16];
    cs.success = values[17];
    cs.source = values[18];
    return cs;
  }

  /**
   * Set the count of the specified counter to zero and blocks until the
   * specified counter reaches the given count.
   * A registered CounterListener still gets the count events.
   * Warning: This method should never be used from the Event Dispatch Thread 
   * (e.g. in a GUI control callback), because the EDT will be blocked and 
   * the step counter no longer reports when the number of steps are reached.
   * @param counterNb the counter number to wait for (1..8)
   * @param count the minimum count until the method returns
   */
  public void waitCount(final int counterNb, final int count)
  {
    if (counterNb < 1 || counterNb > 8 || count <= 0)
      return;

    if (EventQueue.isDispatchThread())
    {
        JOptionPane.showMessageDialog(null, "Blocking Event Dispatch Thread not allowed",
          "Error in calling waitCount()",
          JOptionPane.ERROR_MESSAGE);
        exit();
    }
    counterAry[counterNb - 1].setCount(0);
    saveCounterListeners[counterNb - 1] = counterListeners[counterNb - 1];

    addCounterListener(new CounterListener()
    {
      public void countChanged(CounterEvent evt)
      {
        if (saveCounterListeners[counterNb - 1] != null)
          saveCounterListeners[counterNb - 1].countChanged(evt);

        if (evt.getCount() >= count)
        {
          wakeUp();
        }
      }
    }, counterNb);

    putSleep();
    counterListeners[counterNb - 1] = saveCounterListeners[counterNb - 1];
  }

  private void putSleep()
  {
    synchronized (monitor)
    {
      try
      {
        monitor.wait();
      }
      catch (InterruptedException ex)
      {
      }
    }
  }

  private void wakeUp()
  {
    synchronized (monitor)
    {
      monitor.notifyAll();
    }
  }

  /**
   * Returns the version information.
   * @return the current version number
   */
  public static String getVersion()
  {
    return version;
  }
}
 
